"""
This **module** gives the interface to the package RDKit
"""
try:
    from rdkit import Chem
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "'rdkit' package needed. Maybe you need 'conda install -c rdkit rdkit'") from exc

from . import Xdict
from .namespace import set_global_alternative_names


def assign_to_rdmol(assign, ignore_bond_type=False):
    """
    This **function** is used to convert an Xponge.Assign to a RDKit.rdMol

    :param assign: the Assign instance
    :param ignore_bond_type: set the bond type always to UNSPECIFIED
    :return: the RDKit.rdMol instance
    """
    mol_a = Chem.RWMol()
    for atom in assign.atoms:
        mol_a.AddAtom(Chem.Atom(atom))
    for atom, bonds in assign.bonds.items():
        for aton, n in bonds.items():
            if aton < atom:
                continue
            if ignore_bond_type or n == -1:
                temp_bond = Chem.BondType.UNSPECIFIED
            elif n == 1:
                temp_bond = Chem.BondType.SINGLE
            elif n == 2:
                temp_bond = Chem.BondType.DOUBLE
            elif n == 3:
                temp_bond = Chem.BondType.TRIPLE
            else:
                raise NotImplementedError
            mol_a.AddBond(atom, aton, temp_bond)
    conf = Chem.Conformer(assign.atom_numbers)
    for i in range(assign.atom_numbers):
        conf.SetAtomPosition(i, assign.coordinate[i])
    mol = mol_a.GetMol()
    mol.AddConformer(conf)
    flags = Chem.rdmolops.SanitizeFlags
    Chem.SanitizeMol(mol, flags.SANITIZE_ALL & ~flags.SANITIZE_PROPERTIES)
    return mol


def insert_atom_type_to_rdmol(mol, res, assign, atom_type_dict=None):
    """
    This **function** inserts the atom types in the force field to the RDKit.rdmol instance.
    This is done by setting the isotope of the atom, so it may not be compatible with the other packages.

    :param mol: the RDKit.rdmol instance
    :param res: the Residue instance corresponding to ``mol``
    :param assign: the Assign instance corresponding to ``mol``
    :param atom_type_dict: the dict mapping the atom type to the isotope number
    :return:
    """
    i = 0
    if atom_type_dict is None:
        atom_type_dict = Xdict()
    for a in mol.GetAtoms():
        atom_type = res.name2atom(assign.names[i]).type.name
        if atom_type not in atom_type_dict.keys():
            atom_type_dict[atom_type] = len(atom_type_dict)
        a.SetIsotope(atom_type_dict[atom_type])
        i += 1


def find_equal_atoms(assign):
    """
    This **function** is used to find the chemical equivalent atoms in the molecule

    :param assign: the Assign instance
    :return: a list of equalvalent atom index lists
    """
    mols = []
    canon_smiles = []
    for i in range(len(assign.atoms)):
        mols.append(assign_to_rdmol(assign))
        mols[-1].GetAtoms()[i].SetIsotope(1)
        canon_smiles.append(Chem.MolToSmiles(mols[-1], isomericSmiles=True))
    group = {i: i for i in range(len(assign.atoms))}
    for i in range(len(assign.atoms)):
        if group[i] == i:
            for j in range(i + 1, len(assign.atoms)):
                if canon_smiles[j] == canon_smiles[i]:
                    group[j] = i
    ret = []
    realmap = Xdict()
    for i in group:
        if group[i] == i:
            ret.append([i])
            realmap[i] = len(realmap)
        else:
            ret[realmap[group[i]]].append(i)
    return list(filter(lambda x: len(x) > 1, ret))


set_global_alternative_names()



"""
This **module** is the basic setting for the force field format of harmonic bond
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names, Xdict

# pylint: disable=invalid-name
BondType = Generate_New_Bonded_Force_Type("bond", "1-2", {"k": float, "b": float}, True)

BondType.Set_Property_Unit("k", "energy·distance^-2", "kcal/mol·A^-2")
BondType.Set_Property_Unit("b", "distance", "A")


@Molecule.Set_Save_SPONGE_Input("bond")
def write_bond(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    bonds = []
    for bond in self.bonded_forces.get("bond", []):
        order = list(range(2))
        if bond.k != 0:
            if self.atom_index[bond.atoms[order[0]]] > self.atom_index[bond.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            bonds.append("%d %d %f %f" % (self.atom_index[bond.atoms[temp_order[0]]]
                                          , self.atom_index[bond.atoms[temp_order[1]]], bond.k, bond.b))

    if bonds:
        towrite = "%d\n" % len(bonds)
        bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
        towrite += "\n".join(bonds)

        return towrite
    return None


@Molecule.Set_MindSponge_Todo("bond")
def _do(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    from mindsponge.potential import BondEnergy
    if "bond" not in sys_kwarg:
        sys_kwarg["bond"] = []
    if "bond" not in ene_kwarg:
        ene_kwarg["bond"] = Xdict()
        ene_kwarg["bond"]["function"] = lambda system, ene_kwarg: BondEnergy(
            index=system.bond, use_pbc=use_pbc,
            force_constant=ene_kwarg["bond"]["force_constant"],
            bond_length=ene_kwarg["bond"]["bond_length"],
            length_unit="A", energy_unit="kcal/mol")
        ene_kwarg["bond"]["force_constant"] = []
        ene_kwarg["bond"]["bond_length"] = []
    bonds = []
    force_constants = []
    bond_lengths = []
    for bond in self.bonded_forces.get("bond", []):
        if bond.k == 0:
            continue
        bonds.append([self.atom_index[bond.atoms[0]], self.atom_index[bond.atoms[1]]])
        force_constants.append(bond.k * 2)
        bond_lengths.append(bond.b)
    sys_kwarg["bond"].append(bonds)
    ene_kwarg["bond"]["force_constant"].append(force_constants)
    ene_kwarg["bond"]["bond_length"].append(bond_lengths)


set_global_alternative_names()



"""
This **package** helps to run the molecular dynamics simulation
"""
import os
import sys
from ..helper import Xopen, set_global_alternative_names


def run(args):
    """
    call SPONGE to run the MD simulation

    :param args: a string or a list of arguments
    :return: None
    """
    if isinstance(args, str):
        args = args.split()
        args.insert(0, sys.argv[0])

    if len(args) < 2 or args[1] in ("-h", "--help"):
        print(""" mdrun: run the SPONGE md simulation
        Usage:
            mdrun, mdrun -h, mdrun --help: see this help
            mdrun -set BIN_PATH: set the SPONGE bin direction path to BIN_PATH
                                 BIN_PATH can be an absolute path or a relative path to this module file 
            mdrun SPONGE*:  run SPONGE""")
        sys.exit()

    this_path = os.path.dirname(__file__)

    if args[1] == "-set":
        f = Xopen(os.path.join(this_path, "BIN_PATH.dat"), "w")
        f.write(args[2])
        f.close()
        sys.exit()

    f = Xopen(os.path.join(this_path, "BIN_PATH.dat"), "r")
    that_path = f.read().strip()
    f.close()

    if not os.path.isabs(that_path):
        that_path = os.path.join(this_path, that_path)

    cmd = os.path.join(that_path, args[1])
    if not (os.path.exists(cmd) or os.path.exists(cmd + ".exe")):
        print("Error: No MD Engine found.\n",
              f"  There is no executable program named '{args[1]}' in '{that_path}'\n",
              "  Maybe you need to use Xponge.mdrun -set SPONGE_PATH to set the path to MD Engine",
              file=sys.__stderr__)
        sys.exit(1)
    if len(args) > 2:
        cmd += " " + " ".join(args[2:])

    os.system(cmd)

set_global_alternative_names()



# Copyright 2021 The AIMM Group at Shenzhen Bay Laboratory & Peking University & Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Xponge, a lightweight and easy to customize python package for pre- and post- process of molecular modelling

.. TIP::

    When you use this **package** as an independent package, this package is called ``Xponge``; when you use it as \
a part of mindsponge, it is called ``mindsponge.toolkits``. The documatation is all written as ``Xponge``, but \
remember to change your codes for importing according to your environment!

name style
============

All modules follow the lower snake case such as ``forcefield.amber.ff14sb``

All classes follow the upper camel case such as ``ResidueType``, ``Molecule``, ``GromacsTopologyIterator``.

All functions can be used in four name styles. Take the function ``impose_bond`` as example::

    impose_bond == Impose_Bond == ImposeBond == imposeBond

For some abbreviations, all upper letters are also acceptable. Take the function ``load_pdb`` as example::

    load_pdb == Load_Pdb == Load_PDB == LoadPdb == LoadPDB == loadPdb == loadPDB

Functions in a Python class, no matter it is called a class method, a instance method or a static method follow the \
same name styles as the usual functions do. Take the function ``Residue.add_atom`` as  example::

    Residue.add_atom == Residue.Add_Atom == Residue.addAtom == Residue.AddAtom

name space
============

The submodules in ``forcefield`` will do some global configure settings after loading. For example, after the code::

    import Xponge.forcefield.amber.ff14sb

you can use all ``ResidueType`` instances in ff14sb then.

The names of all ``ResidueType`` instances will be loaded into the main dict, which means you can directly use it::

    import Xponge.forcefield.amber.ff14sb
    print(ALA)
    # output:
    # Type of Residue: ALA

All functions in the base module is also loaded into the main dict, \
which means you can directly use it without the module name. For example, you can use::

    import Xponge
    load_mol2("example.mol2")
    # This is the same as
    # Xponge.load_mol2("example.mol2")

The atoms in a ``Residue`` or a ``ResidueType`` can be obtained by their names. For example::

    import Xponge.forcefield.amber.ff14sb
    print(ALA.CA)

"""

__version__ = "1.2.6.12.0"

import os
import time
import sys
from collections import OrderedDict, deque
from itertools import product, permutations

import numpy as np

from . import assign
from .assign import Assign, get_assignment_from_pdb, get_assignment_from_mol2, get_assignment_from_pubchem, \
    get_assignment_from_residuetype
from .helper import GlobalSetting, Type, AbstractMolecule, ResidueType, Entity, Atom, Residue, ResidueLink, Molecule, \
    AtomType, set_global_alternative_names, generate_new_pairwise_force_type, generate_new_bonded_force_type, source, \
    Xprint
from .load import load_ffitp, load_mol2, load_rst7, load_frcmod, load_pdb, load_parmdat, load_coordinate
from .build import save_mol2, save_pdb, save_sponge_input, save_gro, build_bonded_force, get_mindsponge_system_energy
from .process import impose_bond, impose_angle, impose_dihedral, add_solvent_box, h_mass_repartition, solvent_replace, \
    main_axis_rotate, get_peptide_from_sequence, optimize, Region, UnionRegion, IntersectRegion, \
    BlockRegion, SphereRegion, FrustumRegion, PrismRegion, Lattice


def _initialize():
    """

    :return:
    """
    set_global_alternative_names(True)
    AtomType.New_From_String("name\nUNKNOWN")

    def write_residue(self):
        towrite = "%d %d\n" % (len(self.atoms), len(self.residues))
        towrite += "\n".join([str(len(res.atoms)) for res in self.residues])
        return towrite

    Molecule.Set_Save_SPONGE_Input("residue")(write_residue)

    def write_coordinate(self):
        towrite = "%d\n" % (len(self.atoms))
        boxlength = [0, 0, 0, self.box_angle[0], self.box_angle[1], self.box_angle[2]]
        maxi = [-float("inf"), -float("inf"), -float("inf")]
        mini = [float("inf"), float("inf"), float("inf")]
        for atom in self.atoms:
            if atom.x > maxi[0]:
                maxi[0] = atom.x
            if atom.y > maxi[1]:
                maxi[1] = atom.y
            if atom.z > maxi[2]:
                maxi[2] = atom.z
            if atom.x < mini[0]:
                mini[0] = atom.x
            if atom.y < mini[1]:
                mini[1] = atom.y
            if atom.z < mini[2]:
                mini[2] = atom.z
        if not GlobalSetting.nocenter and self.box_length is None:
            towrite += "\n".join(
                ["%f %f %f" % (atom.x - mini[0] + 3, atom.y - mini[1] + 3, atom.z - mini[2] + 3) for atom in
                 self.atoms])
        else:
            towrite += "\n".join(["%f %f %f" % (atom.x, atom.y, atom.z) for atom in self.atoms])
        if self.box_length is None:
            boxlength[0] = maxi[0] - mini[0] + 6
            boxlength[1] = maxi[1] - mini[1] + 6
            boxlength[2] = maxi[2] - mini[2] + 6
            self.box_length = [boxlength[0], boxlength[1], boxlength[2]]
        else:
            boxlength[0] = self.box_length[0]
            boxlength[1] = self.box_length[1]
            boxlength[2] = self.box_length[2]
        towrite += "\n%f %f %f %f %f %f" % (
            boxlength[0], boxlength[1], boxlength[2], boxlength[3], boxlength[4], boxlength[5])
        return towrite

    Molecule.Set_Save_SPONGE_Input("coordinate")(write_coordinate)

    #pylint: disable=unused-argument
    def _do_initial(self, sys_kwarg, ene_kwarg, use_pbc):
        if self.box_length is not None:
            box_length = self.box_length
        else:
            crd = self.get_atom_coordinates()
            box_length = list(np.max(crd, axis=0) - np.min(crd, axis=0) + 6)
        if "coordinate" not in sys_kwarg:
            sys_kwarg["coordinate"] = [self.get_atom_coordinates().tolist()]
            sys_kwarg["atoms"] = [[atom.name for atom in self.atoms]]
            sys_kwarg["pbc_box"] = [box_length]
        else:
            sys_kwarg["coordinate"].append(self.get_atom_coordinates().tolist())
            sys_kwarg["atoms"].append([atom.name for atom in self.atoms])
            sys_kwarg["pbc_box"].append(box_length)

    Molecule.Set_MindSponge_Todo("coordinate")(_do_initial)

    Xprint("""Reference for Xponge:
  Xia, Y., Gao, Y. Q.
    Xponge: A Python package to perform pre- and post-processing of molecular simulations
    Journal of Open Source Software, 2022 7(77) 4467
    DOI: 10.21105/joss.04467
""")


_initialize()



"""
This **module** is used to save parameters for special minimization
"""
from ...helper import source, set_global_alternative_names

source("....")


def write_zero_mass_for_hydrogen(self):
    """
    This **function** sets the mass of hydrogen and no lj atoms to zero, to freeze them

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d\n" % (len(self.atoms))
    towrite += "\n".join(
        ["%.3f" % (1) if (atom.mass < 3.999 or atom.LJtype == "ZERO_LJ_ATOM" or
                          atom.bad_coordinate) else "%.3f" % (0) for atom in self.atoms])
    return towrite


def write_zero_lj(self):
    """
    This **function** sets all the lj parameters to zero

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d %d\n\n" % (len(self.atoms), 1)
    for _ in range(1):
        for _ in range(1):
            towrite += "%16.7e" % 0 + " "
        towrite += "\n"
    towrite += "\n"

    for _ in range(1):
        for _ in range(1):
            towrite += "%16.7e" % 0 + " "
        towrite += "\n"
    towrite += "\n"
    towrite += "\n".join(["%d" % (0) for atom in self.atoms])
    return towrite


def write_zero_charge(self):
    """
    This **function** sets all the charge parameters to zero

    :param self: the Molecule instance
    :return: the string to write
    """
    towrite = "%d\n" % (len(self.atoms))
    towrite += "\n".join(["%.6f" % (0) for _ in self.atoms])
    return towrite


def save_min_bonded_parameters():
    """
    This **function** saves parameters to only minimize bonded force when saving SPONGE inputs

    :return: None
    """
    Molecule.Set_Save_SPONGE_Input("fake_mass")(write_zero_mass_for_hydrogen)
    Molecule.Set_Save_SPONGE_Input("fake_LJ")(write_zero_lj)
    Molecule.Set_Save_SPONGE_Input("fake_charge")(write_zero_charge)


def do_not_save_min_bonded_parameters():
    """
    This **function** does not save parameters to only minimize bonded force when saving SPONGE inputs

    :return: None
    """
    Molecule.Del_Save_SPONGE_Input("fake_mass")
    Molecule.Del_Save_SPONGE_Input("fake_LJ")
    Molecule.Del_Save_SPONGE_Input("fake_charge")


set_global_alternative_names()



"""
This **package** is used to assign the properties for atoms, residues and molecules
"""
import heapq
from copy import deepcopy
from collections import OrderedDict
from itertools import groupby
import numpy as np
from ..helper import AtomType, ResidueType, Xopen, Xdict, set_real_global_variable, set_attribute_alternative_names, \
    set_global_alternative_names, Guess_Element_From_Mass, Xprint


class AssignRule:
    """
    This **class** is to be the rule to determine the atom type for one atom

    :param name: the name of the rule
    """
    all = Xdict(not_found_message="AssignRule {} not found. Did you import the proper force field?")

    def __init__(self, name, pure_string=False):
        self.name = name
        AssignRule.all[name] = self
        self.rules = OrderedDict()
        self.pure_string = pure_string
        set_attribute_alternative_names(self)

    def add_rule(self, atomtype):
        """
        This **function** is used as a **decorator** to add the atom type - judge function

        :param atomtype: a string or an AtomType instance
        :return: a **function**, which wraps a judge function (receiving the Assign instance and the atom index \
and giving True or False as a result)
        """
        if isinstance(atomtype, str):
            if not self.pure_string:
                atomtype = AtomType.get_type(atomtype)
        elif not isinstance(atomtype, AtomType):
            raise TypeError("atomtype should be a string or AtomType")

        def wrapper(rule_function):
            self.rules[atomtype] = rule_function

        return wrapper


class _RING():
    """
    This **class** is used to help with the ring assignment.
    """
    def __init__(self, atom_list):
        min_index = np.argmin(atom_list)
        self.atoms = atom_list[min_index:] + atom_list[:min_index]
        reverse_list = self.atoms[::-1]
        reverse_list = reverse_list[-1:] + reverse_list[:-1]
        if reverse_list[1] < self.atoms[1]:
            self.atoms = reverse_list
        self.tohash = "-".join(["%d" % atom for atom in self.atoms])
        self.is_pure_aromatic_ring = None
        self.is_pure_aliphatic_ring = None
        self.is_planar_ring = None
        self.out_plane_double_bond = None

    def __repr__(self):
        return self.tohash

    def __hash__(self):
        return hash(self.tohash)

    def __eq__(self, other):
        return isinstance(other, _RING) and self.tohash == other.tohash

    @staticmethod
    def add_rings_basic_marker(assign, rings):
        """

        :param assign:
        :param rings:
        :return:
        """
        for ring in rings:
            for atom in ring.atoms:
                assign.Add_Atom_Marker(atom, "RG")
                assign.Add_Atom_Marker(atom, "RG%d" % len(ring.atoms))

    @staticmethod
    def check_rings_type(assign, rings):
        """

        :param assign:
        :param rings:
        :return:
        """
        for ring in rings:
            ring.check_pure_aromatic(assign)
            ring.check_pure_aliphatic_and_planar(assign)
            ring.check_out_plane_double_bond(assign)

            if not ring.is_pure_aromatic_ring:
                for atom in ring.atoms:
                    if ring.is_pure_aliphatic_ring:
                        assign.Add_Atom_Marker(atom, "AR5")
                    elif ring.is_planar_ring:
                        if ring.out_plane_double_bond:
                            assign.Add_Atom_Marker(atom, "AR3")
                        else:
                            assign.Add_Atom_Marker(atom, "AR2")
                    else:
                        assign.Add_Atom_Marker(atom, "AR4")

    @staticmethod
    def get_rings(assign):
        """

        :param assign:
        :return:
        """
        current_path = []
        current_path_sons = Xdict()
        current_work = []
        have_found_rings = set([])
        for atom0 in range(len(assign.atoms)):
            current_path.append(atom0)
            current_work.extend([[atom, atom0] for atom in assign.bonds[atom0].keys()])
            current_path_sons[atom0] = len(assign.bonds[atom0])
            current_path_father = []
            while current_path:
                work_atom, from_atom = current_work.pop()
                current_path.append(work_atom)
                current_path_father.append(from_atom)
                bond_atom = []
                for atom in assign.bonds[work_atom].keys():
                    if atom != from_atom:
                        try:
                            index = current_path.index(atom)
                            have_found_rings.add(_RING(current_path[index:]))
                        except ValueError:
                            bond_atom.append([atom, work_atom])

                if len(current_path) < 9:
                    current_path_sons[work_atom] = len(bond_atom)
                    current_work.extend(bond_atom)

                else:
                    current_path_sons[work_atom] = 0

                for atom in current_path[::-1]:
                    if current_path_sons[atom] == 0:
                        pop_atom = current_path.pop()
                        current_path_sons.pop(pop_atom)
                        if current_path_father:
                            father_atom = current_path_father.pop()
                            current_path_sons[father_atom] -= 1
        return have_found_rings

    def get_3_neighbors(self):
        """

        :return:
        """
        for i, atom in enumerate(self.atoms):
            yield self.atoms[i - 2], self.atoms[i - 1], atom

    def check_pure_aromatic(self, assign):
        """

        :param assign:
        :return:
        """
        if len(self.atoms) == 6:
            self.is_pure_aromatic_ring = True
            for atom in self.atoms:
                if not assign.Atom_Judge(atom, "C3") and not assign.Atom_Judge(atom, "N2") and not assign.Atom_Judge(
                        atom, "N3"):
                    self.is_pure_aromatic_ring = False
                    break
                if assign.Atom_Judge(atom, "N3"):
                    temp = 0
                    for bonded_atom, bond_order in assign.bonds[atom].items():
                        temp += bond_order
                    if temp == 3:
                        self.is_pure_aromatic_ring = False
                        break
                for bonded_atom, bond_order in assign.bonds[atom].items():
                    if bond_order == 2 and "RG" not in assign.atom_marker[bonded_atom].keys():
                        self.is_pure_aromatic_ring = False
                        break
                if not self.is_pure_aromatic_ring:
                    break
        else:
            self.is_pure_aromatic_ring = False

    def check_pure_aliphatic_and_planar(self, assign):
        """

        :param assign:
        :return:
        """
        self.is_pure_aliphatic_ring = True
        self.is_planar_ring = True
        for atom in self.atoms:
            if self.is_pure_aromatic_ring:
                assign.Add_Atom_Marker(atom, "AR1")
                for i in range(6):
                    assign.Add_Bond_Marker(self.atoms[i - 1], self.atoms[i], "AB")
            if not assign.Atom_Judge(atom, "C4"):
                self.is_pure_aliphatic_ring = False
            if (not assign.Atom_Judge(atom, "C3") and not assign.Atom_Judge(atom, "N2")
                    and not assign.Atom_Judge(atom, "N3") and not assign.Atom_Judge(atom, "O2")
                    and not assign.Atom_Judge(atom, "S2") and not assign.Atom_Judge(atom, "P2")
                    and not assign.Atom_Judge(atom, "P3")):
                self.is_planar_ring = False

    def check_out_plane_double_bond(self, assign):
        """

        :param assign:
        :return:
        """
        self.out_plane_double_bond = False
        for atom in self.atoms:
            for bonded_atom, order in assign.bonds[atom].items():
                if assign.atoms[bonded_atom] != "C" and order == 2 and bonded_atom not in self.atoms:
                    self.out_plane_double_bond = True


class Assign():
    """
    This **class** is used to assign properties for atoms, which is called an "assignment"

    :param name: the name of the molecule
    """
    XX = set("CNOPS")
    XA = set("OS")
    XB = set("NP")
    XC = set(["F", "Cl", "Br", "I"])
    XD = set("SP")
    XE = set(["N", "O", "F", "Cl", "Br", "S", "I"])

    CONNECTIVITY_RADII = {"H": 0.35, "C": 0.73, "N": 0.66, "O": 0.69, "F": 0.68,
                          "P": 1.04, "S": 0.96, "Cl": 0.95, "Br": 1.08, "I": 1.26}

    def __init__(self, name="ASN"):
        self.name = name
        self.atom_numbers = 0
        self.atoms = []
        self.names = []
        self.element_details = []
        self.coordinate = None
        self.charge = None
        self.atom_types = Xdict()
        self.atom_marker = Xdict()
        self.bonds = Xdict()
        self.ar_bonds = Xdict()
        self.am_bonds = Xdict()
        self.bond_marker = Xdict()
        set_attribute_alternative_names(self)

    def add_index_to_name(self):
        """
        This **function** renames the atoms by adding the index to the element name

        :return: None
        """
        for i in range(self.atom_numbers):
            self.names[i] = self.atoms[i] + str(i)

    def atom_judge(self, atom, string):
        """
        This **function** helps judge whether the atom belongs to the mask. For example, "O2" means an oxygen atom \
connected to two other atoms, "N4" means a nitrogen atom connected to four other atoms.

        :param atom: the index of the atom
        :param string: a string mask  of a list of string masks.
        :return:
        """
        assert isinstance(string, (list, str))
        if isinstance(string, str):
            todo = [string]
        else:
            todo = string
        judge = False
        for s in todo:
            element, links = [''.join(list(g)) for k, g in groupby(s, key=lambda x: x.isdigit())]
            if self.atoms[atom] == element and len(self.bonds[atom]) == int(links):
                judge = True
                break
        return judge

    def add_atom(self, element, x, y, z, name="", charge=0.0):
        """
        This **function** adds an atom to the Assign

        :param element: the chemical symbol for the element. "O" - oxygen, "H" - hydrogen for example.
        :param x: the x coordinate
        :param y: the y coordinate
        :param z: the z coordinate
        :param name: the name of the atom
        :param charge: the charge of the atom
        :return: None
        """
        if "." in element:
            element, element_detail = element.split(".")
            element_detail = "." + element_detail
        else:
            element_detail = ""
        self.element_details.append(element_detail)
        self.atoms.append(element)
        self.bonds[self.atom_numbers] = Xdict()
        self.bond_marker[self.atom_numbers] = Xdict()
        self.atom_marker[self.atom_numbers] = Xdict()
        self.atom_types[self.atom_numbers] = None
        self.atom_numbers += 1
        self.names.append(name)
        if self.coordinate is None:
            self.coordinate = np.array([[float(x), float(y), float(z)]])
        else:
            self.coordinate = np.vstack((self.coordinate, np.array([x, y, z])))
        if self.charge is None:
            self.charge = np.array([charge])
        else:
            self.charge = np.hstack((self.charge, np.array([charge])))

    def add_atom_marker(self, atom, marker):
        """
        This **function** adds a marker to an atom

        :param atom: the atom index
        :param marker: the marker
        :return: None
        """
        if marker in self.atom_marker[atom].keys():
            self.atom_marker[atom][marker] += 1
        else:
            self.atom_marker[atom][marker] = 1

    def add_bond(self, atom1, atom2, order=-1):
        """
        This **function** adds a bond to two atoms

        :param atom1: the index of the first atom
        :param atom2: the index of the the second atom
        :param order: the bond order
        :return: None
        """
        self.bonds[atom1][atom2] = order
        self.bond_marker[atom1][atom2] = set([])
        self.bonds[atom2][atom1] = order
        self.bond_marker[atom2][atom1] = set([])

    def add_bond_marker(self, atom1, atom2, marker, only1=False):
        """
        This **function** adds a marker to a bond

        :param atom1: the index of the first atom
        :param atom2: the index of the the second atom
        :param marker: the marker
        :param only1: only add the marker to the atom1 - atom2 bond instead of the atom2 - atom1 bond
        :return: None
        """
        self.bond_marker[atom1][atom2].add(marker)
        if marker in self.atom_marker[atom1]:
            self.atom_marker[atom1][marker] += 1
        else:
            self.atom_marker[atom1][marker] = 1
        if not only1:
            self.bond_marker[atom2][atom1].add(marker)
            if marker in self.atom_marker[atom2]:
                self.atom_marker[atom2][marker] += 1
            else:
                self.atom_marker[atom2][marker] = 1

    def determine_equal_atoms(self):
        """
        This **function** dertermines the chemical equalvalent atoms

        .. NOTE::

            The pyckage RDKit is needed for this **function**

        :return: a list of equalvalent atom index lists
        """
        from ..helper.rdkit import Find_Equal_Atoms
        return Find_Equal_Atoms(self)

    def determine_ring_and_bond_type(self):
        """
        This **function** determine the ring and the bond type

        :return: None
        """
        have_found_rings = _RING.get_rings(self)
        _RING.add_rings_basic_marker(self, have_found_rings)
        _RING.check_rings_type(self, have_found_rings)

        for atom in range(len(self.atoms)):
            dlo = 0
            noto = 0
            for atom2, order in self.bonds[atom].items():
                if self.Atom_Judge(atom2, "O1"):
                    dlo += 1
                else:
                    noto += 1
            if dlo >= 1 >= noto:
                for atom2, order in self.bonds[atom].items():
                    if self.Atom_Judge(atom2, "O1"):
                        self.Add_Bond_Marker(atom, atom2, "DLB")
            for atom2, order in self.bonds[atom].items():
                if "DLB" in self.bond_marker[atom][atom2]:
                    self.Add_Bond_Marker(atom, atom2, "DL", True)
                    self.Add_Bond_Marker(atom, atom2, "sb", True)
                elif order == 1:
                    self.Add_Bond_Marker(atom, atom2, "sb", True)
                    if "AB" not in self.bond_marker[atom][atom2]:
                        self.Add_Bond_Marker(atom, atom2, "SB", True)
                elif order == 2:
                    self.Add_Bond_Marker(atom, atom2, "db", True)
                    if "AB" not in self.bond_marker[atom][atom2]:
                        self.Add_Bond_Marker(atom, atom2, "DB", True)
                else:
                    self.Add_Bond_Marker(atom, atom2, "tb", True)

    def determine_atom_type(self, rule):
        """
        This **function** determines the atom type.

        .. ATTENTION::

            Before determining the atom type, the ring and bond type should be determined. If you use the function \
``Get_Assignment_From_XXX``, the ring and bond type has been determined by the function, but when you build the Assign \
instance yourself, remember to determine the ring and bond type!

        :param rule: a string or an AssignRule instance
        :return: if the attribute "pure_string" of the rule is False, the atom types will be saved inplace and return
None, else return the atom types.
        """
        if isinstance(rule, str):
            rule = AssignRule.all[rule]
        if rule.pure_string:
            backup = deepcopy(self.atom_types)
        for i in range(len(self.atoms)):
            find_type = False
            for atom_type, type_rule in rule.rules.items():
                if type_rule(i, self):
                    self.atom_types[i] = atom_type
                    find_type = True
                    break

            assert find_type, "No atom type found for assignment %s of atom #%d" % (self.name, i)
        if rule.pure_string:
            backup, self.atom_types = self.atom_types, backup
            return backup
        return None

    def determine_connectivity(self, simple_cutoff=None, tolerance=1.00):
        """
        This **function** determines the connectivity based on atomic distances

        :param simple_cutoff: the distance cutoff to determine whether the two atoms are connected. If None (default),
the rule described in the reference (J. Wang et al., J. Am. Chem. Soc, 2001) will be used.
        :param tolerance: the tolerance factor for the default method
        :return: None
        """
        if simple_cutoff is None:
            for i, ci in enumerate(self.coordinate):
                for j in range(i + 1, self.atom_numbers):
                    dij = np.linalg.norm(np.array(self.coordinate[j]) - np.array(ci))
                    rij = self.CONNECTIVITY_RADII[self.atoms[i]] + self.CONNECTIVITY_RADII[self.atoms[j]]
                    if dij <= 1.5:
                        factor = 1 - 0.15
                    elif dij <= 1.9:
                        factor = 1 - 0.11
                    elif dij <= 2.05:
                        factor = 1 - 0.09
                    else:
                        factor = 1 - 0.08
                    factor /= tolerance
                    if factor < rij / dij < 2:
                        self.add_bond(i, j, -1)
        else:
            for i, ci in enumerate(self.coordinate):
                for j in range(i + 1, self.atom_numbers):
                    dij = np.linalg.norm(np.array(self.coordinate[j]) - np.array(ci))
                    if dij < simple_cutoff:
                        self.add_bond(i, j, -1)

    def determine_bond_order(self, max_step=200, max_stat=2000, penalty_scores=None, debug=False):
        """
        This **function** determines the bond order based on connectivities

        :param max_step: the max iterative step
        :param max_stat: the max iterative stat
        :param penalty_scores: the penalty scores for every atom. This should be a list of ordered dicts, and every
ordered dict stores the valence-penalty pairs for every atom, and it is sorted by the penalty scores. If None(default),
a set of penalty scores described in the reference (J. Wang et al., J. Mol. Graph. Model., 2006) will be used.
        :return: True for success, False for failure.
        """
        from .bond_order import BondOrderAssignment
        if penalty_scores is None:
            penalty_scores = [BondOrderAssignment.atomic_valence[type_]
                              for type_ in self.determine_atom_type("bo").values()]
        bo_assign = BondOrderAssignment(penalty_scores, max_step, max_stat, self, debug)
        return bo_assign.main()

    def to_residuetype(self, name, charge=None):
        """
        This **function** converts the Assign instance to the ResidueType instance

        :param name: the name of the ResidueType instance
        :param charge: the charge of atoms. If set to None, internal charge will be used
        :return: the ResidueType instance
        """
        temp = ResidueType(name=name)
        if not charge:
            if self.charge is None:
                charge = np.zeros(self.atom_numbers)
            else:
                charge = self.charge
        count = Xdict()
        for i in range(self.atom_numbers):
            assert self.atom_types[i] is not None
            if self.names[i]:
                atom_name = self.names[i]
            elif self.atoms[i] in count.keys():
                atom_name = self.atoms[i] + "%d" % count[self.atoms[i]]
                self.names[i] = atom_name
                count[self.atoms[i]] += 1
            else:
                count[self.atoms[i]] = 1
                atom_name = self.atoms[i]
                self.names[i] = atom_name
            temp.Add_Atom(atom_name, self.atom_types[i], x=self.coordinate[i][0],
                          y=self.coordinate[i][1], z=self.coordinate[i][2])
            temp.atoms[-1].charge = charge[i]
        for i, bondi in self.bonds.items():
            for j in bondi.keys():
                temp.Add_Connectivity(temp.atoms[i], temp.atoms[j])
        set_real_global_variable(name, temp)
        return temp

    def calculate_charge(self, method, **parameters):
        """
        This **function** calculates the partial charge for every atom.
        the method "RESP" to calculate the partial charge is not available on Windows.

        :param method: the method to calculate the charge
        :param parameters: the parameters to calculate the charge
        :return: None
        """
        method = method.upper()
        if method == "RESP":
            from . import resp
            self.charge = resp.RESP_Fit(self, basis=parameters.get("basis", "6-31g*"), opt=parameters.get("opt", False),
                                        charge=parameters.get("charge", 0), spin=parameters.get("spin", 0),
                                        extra_equivalence=parameters.get("extra_equivalence", []),
                                        grid_density=parameters.get("grid_density", 6),
                                        grid_cell_layer=parameters.get("grid_cell_layer", 4),
                                        a1=parameters.get("a1", 0.0005),
                                        a2=parameters.get("a2", 0.001), two_stage=parameters.get("two_stage", True),
                                        only_esp=parameters.get("only_esp", False),
                                        radius=parameters.get("radius", None))
        elif method == "GASTEIGER":
            from . import gasteiger
            self.charge = gasteiger.Gasteiger(self)
        else:
            raise ValueError("methods should be one of the following: 'RESP', 'GASTEIGER'")

    def save_as_pdb(self, filename):
        """
        This **function** saves the instance as a pdb file

        :param filename: the name of the output file
        :return: None
        """
        if not isinstance(filename, str):
            raise TypeError("filename needed to save an assignment to a pdb file")
        towrite = towrite = "REMARK   Generated By Xponge (Assignment)\n"
        count = Xdict()
        for i in range(self.atom_numbers):
            if self.names[i]:
                atom_name = self.names[i]
            elif self.atoms[i] in count.keys():
                atom_name = self.atoms[i] + "%d" % count[self.atoms[i]]
                self.names[i] = atom_name
                count[self.atoms[i]] += 1
            else:
                count[self.atoms[i]] = 1
                atom_name = self.atoms[i]
                self.names[i] = atom_name
        for i, atom in enumerate(self.atoms):
            towrite += "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f%17s%2s\n" % (i + 1, self.names[i],
                                                                                 self.name, " ", 1,
                                                                                 self.coordinate[i][0],
                                                                                 self.coordinate[i][1],
                                                                                 self.coordinate[i][2], " ", atom)

        for i in range(self.atom_numbers):
            bonded_atoms = list(self.bonds[i].keys())
            bonded_atoms.sort()
            bonded_atoms = [bonded_atoms[i:i + 4] for i in range(0, len(bonded_atoms), 4)]
            if bonded_atoms:
                for atoms in bonded_atoms:
                    towrite += "CONECT %4d" % (i + 1)
                    for atom in atoms:
                        towrite += " %4d" % (atom + 1)
                    towrite += "\n"

        f = Xopen(filename, "w")
        f.write(towrite)
        f.close()

    def save_as_mol2(self, filename):
        """
        This **function** saves the instance as a mol2 file

        :param filename: the name of the output file
        :return: None
        """
        if not isinstance(filename, str):
            raise TypeError("filename needed to save an assignment as a mol2 file")
        bonds = []
        for i in range(self.atom_numbers):
            for j, order in self.bonds[i].items():
                if i < j:
                    if i in self.ar_bonds.keys() and j in self.ar_bonds[i]:
                        bonds.append("%6d %6d ar\n" % (i + 1, j + 1))
                    elif i in self.am_bonds.keys() and j in self.am_bonds[i]:
                        bonds.append("%6d %6d am\n" % (i + 1, j + 1))
                    else:
                        bonds.append("%6d %6d %1d\n" % (i + 1, j + 1, order))
        bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
        count = Xdict()
        for i in range(self.atom_numbers):
            if self.names[i]:
                atom_name = self.names[i]
            elif self.atoms[i] in count.keys():
                atom_name = self.atoms[i] + "%d" % count[self.atoms[i]]
                self.names[i] = atom_name
                count[self.atoms[i]] += 1
            else:
                count[self.atoms[i]] = 1
                atom_name = self.atoms[i]
                self.names[i] = atom_name
        towrite = "@<TRIPOS>MOLECULE\n%s\n %d %d 1 0 1\nSMALL\nUSER_CHARGES\n" % (
            self.name, self.atom_numbers, len(bonds))
        towrite += "@<TRIPOS>ATOM\n"
        for i, atom in enumerate(self.atoms):
            towrite += "%6d %4s %8.4f %8.4f %8.4f   %-8s %5d %8s %10.6f\n" % (
                i + 1, self.names[i], self.coordinate[i][0], self.coordinate[i][1], self.coordinate[i][2],
                atom + self.element_details[i], 1, self.name, self.charge[i])

        towrite += "@<TRIPOS>BOND\n"
        for i, bond in enumerate(bonds):
            towrite += "%6d %s" % (i + 1, bond)
        towrite += "@<TRIPOS>SUBSTRUCTURE\n"
        towrite += "%5d %8s %6d ****               0 ****  **** \n" % (1, self.name, 1)

        f = Xopen(filename, "w")
        f.write(towrite)
        f.close()


def get_assignment_from_pubchem(parameter, keyword):
    """
    This **function** gets an Assign instance from PubChem

    usage example::

        a1 = Get_Assignment_From_PubChem("ethane", "name")
        a2 = Get_Assignment_From_PubChem("CC", "smiles")

    :param parameter: the parameter to search on PubChem
    :param keyword: the keyword to search on PubChem
    :return: the Assign instance
    """
    import pubchempy as pcp
    cs = pcp.get_compounds(parameter, keyword, record_type='3d')
    if not cs:
        raise pcp.NotFoundError
    if len(cs) == 1:
        assign = Assign()
        c = cs[0]
        for atom in c.atoms:
            assign.Add_Atom(atom.element, atom.x, atom.y, atom.z)
        for bond in c.bonds:
            assign.Add_Bond(bond.aid1 - 1, bond.aid2 - 1, bond.order)
        assign.Determine_Ring_And_Bond_Type()
        return assign
    raise NotImplementedError


def get_assignment_from_pdb(filename, determine_bond_order=True, only_residue=""):
    """
    This **function** gets an Assign instance from a pdb file

    :param filename: the name of the input file
    :param determine_bond_order: whether determine the bond order automatically
    :param only_residue: only get the residue with the name same as ``only_residue``
    :return: the Assign instance
    :raise NotImplementedError: it has not been implemented when determine_bond_order is True
    """
    assign = Assign()
    index_atom_map = Xdict()
    with open(filename) as f:
        for line in f:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                if only_residue:
                    resname = line[17:20].strip()
                    if resname != only_residue:
                        continue
                index = int(line[6:11])
                index_atom_map[index] = assign.atom_numbers
                atom_name = line[12:16].strip()
                element = line[76:78].strip()
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                assign.Add_Atom(element, x, y, z, atom_name)
            if line.startswith("CONECT"):
                atom = int(line[6:11])
                if atom not in index_atom_map.keys():
                    continue
                for bonded_atom_i in range(11, 31, 5):
                    try:
                        temp = line[bonded_atom_i:bonded_atom_i + 5]
                        bonded_atom = int(temp)
                    except ValueError:
                        break
                    if bonded_atom in index_atom_map.keys():
                        assign.Add_Bond(index_atom_map[atom], index_atom_map[int(bonded_atom)])
    if determine_bond_order:
        raise NotImplementedError
    return assign


def get_assignment_from_residuetype(restype):
    """
    This **function** gets an Assign instance from a ResidueType instance

    :param restype: the ResidueType instance
    :return: the Assign instance
    """
    assign = Assign()
    for atom in restype.atoms:
        assign.Add_Atom(Guess_Element_From_Mass(atom.mass), atom.x, atom.y, atom.z, atom.name)
    for atom in restype.atoms:
        i = restype.atom2index(atom)
        for atomb in restype.connectivity[atom]:
            j = restype.atom2index(atomb)
            if i < j:
                assign.Add_Bond(i, j)
    return assign


def _deal_with_ar_bonds(assign):
    """

    :param assign:
    :return:
    """
    ar_bonds_atoms = list(assign.ar_bonds.keys())
    ar_bonds_atoms.sort(key=lambda x: (x, len(assign.ar_bonds[x])))
    doubled = Xdict()
    checked = Xdict()
    for ar_atom in ar_bonds_atoms:
        assign.ar_bonds[ar_atom].sort(key=lambda x: (x, len(assign.ar_bonds[x])))
        doubled[ar_atom] = False
        checked[ar_atom] = False

    working_space = []
    while ar_bonds_atoms:
        working_space.append(ar_bonds_atoms.pop())
        while working_space:
            work_atom = working_space.pop()
            if checked[work_atom]:
                continue
            checked[work_atom] = True
            for neighbor in assign.ar_bonds[work_atom]:
                if not checked[neighbor]:
                    working_space.append(neighbor)
            for neighbor in assign.ar_bonds[work_atom][::-1]:
                if doubled[work_atom]:
                    break
                if not doubled[neighbor] and not doubled[work_atom]:
                    assign.bonds[neighbor][work_atom] = 2
                    doubled[neighbor] = True
                    doubled[work_atom] = True


def get_assignment_from_mol2(filename, ignore_hydrogen=False):
    """
    This **function** gets an Assign instance from a mol2 file

    :param filename: the name of the input file
    :return: the Assign instance
    """
    with open(filename) as f:
        flag = None
        subflag = None
        real_index = Xdict()
        for line in f:
            if not line.strip():
                continue
            if line.startswith("@<TRIPOS>"):
                flag = line[9:].strip()
            elif flag == "MOLECULE":
                if subflag is None:
                    assign = Assign(line.strip())
                    subflag = "0"
            # 处理原子信息
            elif flag == "ATOM":
                words = line.split()
                atom_name = words[1]
                if ignore_hydrogen and (atom_name.startswith("H") or atom_name.startswith("1H")  or
                    atom_name.startswith("2H") or atom_name.startswith("3H")):
                    continue
                real_index[words[0]] = assign.atom_numbers
                element = words[5]
                x = float(words[2])
                y = float(words[3])
                z = float(words[4])
                charge = float(words[8])
                assign.Add_Atom(element, x, y, z, atom_name, charge)
            elif flag == "BOND":
                words = line.split()
                if words[1] not in real_index or words[2] not in real_index:
                    continue
                atom1 = real_index[words[1]]
                atom2 = real_index[words[2]]
                if words[3] in "1234567890":
                    assign.Add_Bond(atom1 , atom2, int(words[3]))
                elif words[3] == "ar":
                    assign.Add_Bond(atom1, atom2, 1)
                    if atom1 not in assign.ar_bonds.keys():
                        assign.ar_bonds[atom1] = [atom2]
                    else:
                        assign.ar_bonds[atom1].append(atom2)
                    if atom2 not in assign.ar_bonds.keys():
                        assign.ar_bonds[atom2] = [atom1]
                    else:
                        assign.ar_bonds[atom2].append(atom1)
                elif words[3] == "am":
                    assign.Add_Bond(atom1, atom2, 1)
                    if atom1 not in assign.am_bonds.keys():
                        assign.am_bonds[atom1] = [atom2]
                    else:
                        assign.am_bonds[atom1].append(atom2)
                    if atom2 not in assign.am_bonds.keys():
                        assign.am_bonds[atom2] = [atom1]
                    else:
                        assign.am_bonds[atom2].append(atom1)
                else:
                    raise NotImplementedError(f"No implemented method to process bond #{words[0]} type {words[3]}")

    _deal_with_ar_bonds(assign)
    assign.Determine_Ring_And_Bond_Type()
    return assign

set_global_alternative_names()



"""
This **module** is the basic setting for the force field format of Urey Bradley angle
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names

# pylint: disable=invalid-name
UreyBradleyType = Generate_New_Bonded_Force_Type("Urey_Bradley", "1-2-3",
                                                 {"k": float, "b": float, "kUB": float, "r13": float}, True)

UreyBradleyType.Set_Property_Unit("k", "energy·angle^-2", "kcal/mol·rad^-2")
UreyBradleyType.Set_Property_Unit("b", "angle", "rad")
UreyBradleyType.Set_Property_Unit("kUB", "energy·distance^-2", "kcal/mol·A^-2")
UreyBradleyType.Set_Property_Unit("r13", "distance", "A")


@Molecule.Set_Save_SPONGE_Input("urey_bradley")
def write_angle(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    angles = []
    for angle in self.bonded_forces.get("Urey_Bradley", []):
        order = list(range(3))
        if angle.k != 0 or angle.kUB != 0:
            if self.atom_index[angle.atoms[order[0]]] > self.atom_index[angle.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            angles.append("%d %d %d %f %f %f %f" % (self.atom_index[angle.atoms[temp_order[0]]]
                                                    , self.atom_index[angle.atoms[temp_order[1]]],
                                                    self.atom_index[angle.atoms[temp_order[2]]], angle.k, angle.b,
                                                    angle.kUB, angle.r13))

    if angles:
        towrite = "%d\n" % len(angles)
        angles.sort(key=lambda x: list(map(int, x.split()[:3])))
        towrite += "\n".join(angles)
        return towrite
    return None

set_global_alternative_names()



"""
This **module** is the basic setting for the force field format of 2-parameter non bonded 1-4 interactions
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names, Xdict

# pylint: disable=invalid-name
NB14Type = Generate_New_Bonded_Force_Type("nb14", "1-4", {"kLJ": float, "kee": float}, True)

NB14Type.topology_matrix = [[1, -4],
                            [1, 1]]


@Molecule.Set_Save_SPONGE_Input("nb14")
def write_nb14(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    bonds = []
    for bond in self.bonded_forces.get("nb14", []):
        order = list(range(2))
        if bond.kLJ != 0 and bond.kee != 0:
            if self.atom_index[bond.atoms[order[0]]] > self.atom_index[bond.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            bonds.append("%d %d %f %f" % (self.atom_index[bond.atoms[temp_order[0]]]
                                          , self.atom_index[bond.atoms[temp_order[1]]], bond.kLJ, bond.kee))

    if bonds:
        towrite = "%d\n" % len(bonds)
        bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
        towrite += "\n".join(bonds)

        return towrite
    return None


#pylint: disable=unused-argument, relative-beyond-top-level
@Molecule.Set_MindSponge_Todo("nb14")
def _do(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    from mindsponge.potential import NonbondPairwiseEnergy
    from .nb14_extra_base import get_nb14_extra_lj
    if "nb14" not in ene_kwarg:
        ene_kwarg["nb14"] = Xdict()
        ene_kwarg["nb14"]["function"] = lambda sys_kwarg, ene_kwarg: NonbondPairwiseEnergy(
            index=ene_kwarg["nb14"]["index"], use_pbc=use_pbc,
            qiqj=ene_kwarg["nb14"]["qiqj"],
            epsilon_ij=ene_kwarg["nb14"]["epsilon_ij"],
            sigma_ij=ene_kwarg["nb14"]["sigma_ij"],
            r_scale=ene_kwarg["nb14"]["r_scale"],
            r6_scale=ene_kwarg["nb14"]["r6_scale"],
            r12_scale=ene_kwarg["nb14"]["r12_scale"],
            length_unit="A", energy_unit="kcal/mol")
        ene_kwarg["nb14"]["index"] = []
        ene_kwarg["nb14"]["r_scale"] = []
        ene_kwarg["nb14"]["r6_scale"] = []
        ene_kwarg["nb14"]["r12_scale"] = []
        ene_kwarg["nb14"]["qiqj"] = []
        ene_kwarg["nb14"]["sigma_ij"] = []
        ene_kwarg["nb14"]["epsilon_ij"] = []
    nb14s = []
    r_scales = []
    r6_scales = []
    r12_scales = []
    qiqj = []
    sigma_ij = []
    epsilon_ij = []
    for bond in self.bonded_forces.get("nb14", []):
        if bond.kee != 1 and bond.kLJ != 1:
            atom1, atom2 = bond.atoms
            a, b = get_nb14_extra_lj(atom1, atom2)
            nb14s.append([self.atom_index[atom] for atom in bond.atoms])
            qiqj.append(atom1.charge * atom2.charge)
            if a != 0:
                sigma_ij.append((a/b)**(1/6))
                epsilon_ij.append(b*b/a/4)
            else:
                sigma_ij.append(0)
                epsilon_ij.append(0)
            r_scales.append(bond.kee)
            r6_scales.append(bond.kLJ)
            r12_scales.append(bond.kLJ)
    ene_kwarg["nb14"]["index"].append(nb14s)
    ene_kwarg["nb14"]["r_scale"].append(r_scales)
    ene_kwarg["nb14"]["r6_scale"].append(r6_scales)
    ene_kwarg["nb14"]["r12_scale"].append(r12_scales)
    ene_kwarg["nb14"]["qiqj"].append(qiqj)
    ene_kwarg["nb14"]["sigma_ij"].append(sigma_ij)
    ene_kwarg["nb14"]["epsilon_ij"].append(epsilon_ij)

set_global_alternative_names()



"""
This **package** gives functions and classes for analysis
"""
import numpy as np


class MdoutReader():
    """
    This **class** is used to read the mdout file generated by SPONGE

    Suppose we have a mdout file named "example.mdout" like this::

        step    Time    Temperature     Potential   LJ
        0       0       300.0           100.0       100.0
        1       0.1     299.8           101.1       101.1

    Then we can use the following codes to read it::

        from Xponge.analysis import MdoutReader
        mdout = MdoutReader("example.mdout")
        print(mdout.step, mdout.Time, mdout.Temperature, mdout.Potential, mdout.LJ)

    :param filename: the name of the mdout file
    """
    def __init__(self, filename):
        with open(filename) as f:
            self.content = f.readline().split()
            words = f.read().replace("****", "nan").split()
            self.data = np.array([[float(words[i+j])
                                   for j in range(len(self.content))]
                                  for i in range(0, len(words), len(self.content))])
        self.content_index = {self.content[i]: i for i in range(len(self.content))}

    def __getattribute__(self, attr):
        if attr not in ("content", "content_index", "data") and attr in self.content_index.keys():
            return self.data[:, self.content_index[attr]]
        return super().__getattribute__(attr)



"""
This **module** is the basic setting for the force field format of periodic proper and improper dihedral
"""
from itertools import permutations
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names, Xdict

# pylint: disable=invalid-name
ProperType = Generate_New_Bonded_Force_Type("dihedral", "1-2-3-4", {"k": float, "phi0": float, "periodicity": int},
                                            True, ["k", "phi0", "periodicity"])

ProperType.Set_Property_Unit("k", "energy", "kcal/mol")
ProperType.Set_Property_Unit("phi0", "angle", "rad")

# pylint: disable=invalid-name
ImproperType = Generate_New_Bonded_Force_Type("improper", "1-3-2-3", {"k": float, "phi0": float, "periodicity": int},
                                              False)
ImproperType.topology_matrix = [[1, 3, 2, 3],
                                [1, 1, 2, 3],
                                [1, 1, 1, 2],
                                [1, 1, 1, 1]]

ImproperType.Set_Property_Unit("k", "energy", "kcal/mol")
ImproperType.Set_Property_Unit("phi0", "angle", "rad")


@ImproperType.Set_Same_Force_Function
def improper_same_force(_, atom_list):
    """
    This **function** is used to return the same force type for an atom list
    :param _:
    :param atom_list:
    :return:
    """
    temp = []
    if isinstance(atom_list, str):
        atom_list_temp = [atom.strip() for atom in atom_list.split("-")]
        center_atom = atom_list_temp.pop(2)
        for atom_permutation in permutations(atom_list_temp):
            atom_permutation = list(atom_permutation)
            atom_permutation.insert(2, center_atom)
            temp.append("-".join(atom_permutation))
    else:
        atom_list_temp = list(atom_list)
        center_atom = atom_list_temp.pop(2)
        for atom_permutation in permutations(atom_list_temp):
            atom_permutation = list(atom_permutation)
            atom_permutation.insert(2, center_atom)
            temp.append(atom_permutation)
    return temp


@Molecule.Set_Save_SPONGE_Input("dihedral")
def write_dihedral(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    dihedrals = []
    for dihedral in self.bonded_forces.get("dihedral", []):
        order = list(range(4))
        if self.atom_index[dihedral.atoms[order[0]]] > self.atom_index[dihedral.atoms[order[-1]]]:
            temp_order = order[::-1]
        else:
            temp_order = order
        for i in range(dihedral.multiple_numbers):
            if dihedral.ks[i] != 0:
                dihedrals.append("%d %d %d %d %d %f %f" % (self.atom_index[dihedral.atoms[temp_order[0]]]
                                                           , self.atom_index[dihedral.atoms[temp_order[1]]],
                                                           self.atom_index[dihedral.atoms[temp_order[2]]]
                                                           , self.atom_index[dihedral.atoms[temp_order[3]]],
                                                           dihedral.periodicitys[i], dihedral.ks[i], dihedral.phi0s[i]))

    for dihedral in self.bonded_forces.get("improper", []):
        order = list(range(4))
        if dihedral.k != 0:
            if self.atom_index[dihedral.atoms[order[0]]] > self.atom_index[dihedral.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            dihedrals.append("%d %d %d %d %d %f %f" % (self.atom_index[dihedral.atoms[temp_order[0]]]
                                                       , self.atom_index[dihedral.atoms[temp_order[1]]],
                                                       self.atom_index[dihedral.atoms[temp_order[2]]]
                                                       , self.atom_index[dihedral.atoms[temp_order[3]]],
                                                       dihedral.periodicity, dihedral.k, dihedral.phi0))

    if dihedrals:
        towrite = "%d\n" % len(dihedrals)
        dihedrals.sort(key=lambda x: list(map(float, x.split())))
        towrite += "\n".join(dihedrals)

        return towrite
    return None


#pylint: disable=unused-argument
@Molecule.Set_MindSponge_Todo("dihedral")
def _do(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    from mindsponge.potential import DihedralEnergy
    if "dihedral" not in ene_kwarg:
        ene_kwarg["dihedral"] = Xdict()
        ene_kwarg["dihedral"]["function"] = lambda system, ene_kwarg: DihedralEnergy(
            index=ene_kwarg["dihedral"]["index"], use_pbc=use_pbc,
            force_constant=ene_kwarg["dihedral"]["force_constant"],
            periodicity=ene_kwarg["dihedral"]["periodicity"],
            phase=ene_kwarg["dihedral"]["phase"],
            energy_unit="kcal/mol")
        ene_kwarg["dihedral"]["index"] = []
        ene_kwarg["dihedral"]["force_constant"] = []
        ene_kwarg["dihedral"]["periodicity"] = []
        ene_kwarg["dihedral"]["phase"] = []
    dihedrals = []
    force_constants = []
    periodicitys = []
    phases = []
    for dihedral in self.bonded_forces.get("dihedral", []):
        for i in range(dihedral.multiple_numbers):
            if dihedral.ks[i] != 0:
                dihedrals.append([self.atom_index[atom] for atom in dihedral.atoms])
                force_constants.append(dihedral.ks[i] * 2)
                periodicitys.append(dihedral.periodicitys[i])
                phases.append(dihedral.phi0s[i])
    for dihedral in self.bonded_forces.get("improper", []):
        if dihedral.k != 0:
            dihedrals.append([self.atom_index[atom] for atom in dihedral.atoms])
            force_constants.append(dihedral.k * 2)
            periodicitys.append(dihedral.periodicity)
            phases.append(dihedral.phi0)
    ene_kwarg["dihedral"]["index"].append(dihedrals)
    ene_kwarg["dihedral"]["force_constant"].append(force_constants)
    ene_kwarg["dihedral"]["periodicity"].append(periodicitys)
    ene_kwarg["dihedral"]["phase"].append(phases)

set_global_alternative_names()



"""
This **module** set the basic configuration for gaff
"""
from ...helper import source, Xprint, set_real_global_variable

source("....")
amber = source("...amber")

amber.load_parameters_from_parmdat("gaff.dat")

gaff = assign.AssignRule("gaff")


@gaff.Add_Rule("cx")
def _rule_cx(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C4") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("cy")
def _rule_cy(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C4") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("c3")
def _rule_c3(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C4")


@gaff.Add_Rule("c")
def _rule_c(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3")
    if tofind:
        tofind = False
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, ["O1", "S1"]):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("cz")
def _rule_cz(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3")
    if tofind:
        for bonded_atom in assign.bonds[i].keys():
            if not assign.Atom_Judge(bonded_atom, "N3"):
                tofind = False
                break
    return tofind


@gaff.Add_Rule("cq")
def _rule_cq(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3") and "AR1" in assign.atom_marker[i] and assign.atom_marker[i].get("RG6") == 1
    for bonded_atom in assign.bonds[i].keys():
        if not tofind:
            break
        if assign.atoms[bonded_atom] not in assign.XX or "AR1" not in assign.atom_marker[bonded_atom].keys():
            tofind = False
    if tofind:
        tofind = False
        for bonded_atom in assign.bonds[i]:
            if tofind:
                break
            if (assign.atom_types[bonded_atom] == AtomType.get_type("cp") and
                    "AB" in assign.bond_marker[bonded_atom][i]):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("cp")
def _rule_cp(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3") and "AR1" in assign.atom_marker[i] and assign.atom_marker[i].get("RG6") == 1
    for bonded_atom in assign.bonds[i].keys():
        if not tofind:
            break
        if assign.atoms[bonded_atom] not in assign.XX or "AR1" not in assign.atom_marker[bonded_atom].keys():
            tofind = False

    return tofind


@gaff.Add_Rule("ca")
def _rule_ca(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C3") and "AR1" in assign.atom_marker[i]


def _single_double_name(i, assign, name1, name2):
    """

    :param i:
    :param assign:
    :param name1:
    :param name2:
    :return:
    """
    tofind = False
    for bonded_atom, bond_order in assign.bonds[i].items():
        if tofind:
            break
        if ((assign.atom_types[bonded_atom] == AtomType.get_type(name1) and bond_order == 2) or
                (assign.atom_types[bonded_atom] == AtomType.get_type(name2) and bond_order == 1)):
            tofind = True
            break
    return tofind


@gaff.Add_Rule("cd")
def _rule_cd(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
             ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])
    if tofind:
        tofind = _single_double_name(i, assign, "cc", "cd")
    return tofind


@gaff.Add_Rule("cc")
def _rule_cc(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C3") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
           ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])


@gaff.Add_Rule("cf")
def _rule_cf(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if tofind:
                break
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    if tofind:
        tofind = _single_double_name(i, assign, "ce", "cf")
    return tofind


@gaff.Add_Rule("ce")
def _rule_ce(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C3") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if tofind:
                break
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("cu")
def _rule_cu(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C3") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("cv")
def _rule_cv(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C3") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("c2")
def _rule_c2(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C3")


@gaff.Add_Rule("cg")
def _rule_cg(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "C2") and "sb" in assign.atom_marker[i] and "tb" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if tofind:
                break
            if bond_order == 1 and assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2", "N1"]):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("c1")
def _rule_c1(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "C2") or assign.Atom_Judge(i, "C1")


@gaff.Add_Rule("hn")
def _rule_hn(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1") and assign.atoms[list(assign.bonds[i].keys())[0]] == "N"


@gaff.Add_Rule("ho")
def _rule_ho(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1") and assign.atoms[list(assign.bonds[i].keys())[0]] == "O"


@gaff.Add_Rule("hs")
def _rule_hs(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1") and assign.atoms[list(assign.bonds[i].keys())[0]] == "S"


@gaff.Add_Rule("hp")
def _rule_hp(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1") and assign.atoms[list(assign.bonds[i].keys())[0]] == "P"


@gaff.Add_Rule("hx")
def _rule_hx(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    for bonded_atom in assign.bonds[i].keys():
        if assign.atoms[bonded_atom] == "C":
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.Atom_Judge(bonded_atom_bonded, "N4"):
                    tofind = True
                    break
    return assign.Atom_Judge(i, "H1") and tofind


@gaff.Add_Rule("hw")
def _rule_hw(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    for bonded_atom in assign.bonds[i].keys():
        if assign.atoms[bonded_atom] == "O":
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.Atom_Judge(bonded_atom_bonded, "H1"):
                    tofind = True
                    break
    return assign.Atom_Judge(i, "H1") and tofind


@gaff.Add_Rule("h3")
def _rule_h3(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    for bonded_atom in assign.bonds[i].keys():
        if assign.Atom_Judge(bonded_atom, "C4"):
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.atoms[bonded_atom_bonded] in assign.XE:
                    tofind += 1

    return assign.Atom_Judge(i, "H1") and tofind == 3


@gaff.Add_Rule("h2")
def _rule_h2(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    for bonded_atom in assign.bonds[i].keys():
        if assign.Atom_Judge(bonded_atom, "C4"):
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.atoms[bonded_atom_bonded] in assign.XE:
                    tofind += 1
    return assign.Atom_Judge(i, "H1") and tofind == 2


@gaff.Add_Rule("h1")
def _rule_h1(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    for bonded_atom in assign.bonds[i].keys():
        if assign.Atom_Judge(bonded_atom, "C4"):
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.atoms[bonded_atom_bonded] in assign.XE:
                    tofind += 1
    return assign.Atom_Judge(i, "H1") and tofind == 1


@gaff.Add_Rule("hc")
def _rule_hc(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1") and assign.Atom_Judge(list(assign.bonds[i].keys())[0], "C4")


@gaff.Add_Rule("h5")
def _rule_h5(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    for bonded_atom in assign.bonds[i].keys():
        if assign.Atom_Judge(bonded_atom, "C3"):
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.atoms[bonded_atom_bonded] in assign.XE:
                    tofind += 1
    return assign.Atom_Judge(i, "H1") and tofind == 2


@gaff.Add_Rule("h4")
def _rule_h4(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    for bonded_atom in assign.bonds[i].keys():
        if assign.Atom_Judge(bonded_atom, "C3"):
            for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                if assign.atoms[bonded_atom_bonded] in assign.XE:
                    tofind += 1
    return assign.Atom_Judge(i, "H1") and tofind == 1


@gaff.Add_Rule("ha")
def _rule_ha(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "H1")


@gaff.Add_Rule("f")
def _rule_f(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.atoms[i] == "F"


@gaff.Add_Rule("cl")
def _rule_cl(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.atoms[i] == "Cl"


@gaff.Add_Rule("br")
def _rule_br(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.atoms[i] == "Br"


@gaff.Add_Rule("i")
def _rule_i(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.atoms[i] == "I"


@gaff.Add_Rule("o")
def _rule_o(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "O1")


@gaff.Add_Rule("oh")
def _rule_oh(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "O2") or assign.Atom_Judge(i, "O3"):
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "H1"):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("op")
def _rule_op(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "O2") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("oq")
def _rule_oq(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "O2") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("os")
def _rule_os(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.atoms[i] == "O"


@gaff.Add_Rule("ni")
def _rule_ni(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3") and "RG3" in assign.atom_marker[i].keys():
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "C3"):
                for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                    if assign.Atom_Judge(bonded_atom_bonded, "O1") or assign.Atom_Judge(bonded_atom_bonded, "S1"):
                        tofind = True
                        break
            if tofind:
                break
    return tofind


@gaff.Add_Rule("nj")
def _rule_nj(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3") and "RG4" in assign.atom_marker[i].keys():
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "C3"):
                for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                    if assign.Atom_Judge(bonded_atom_bonded, "O1") or assign.Atom_Judge(bonded_atom_bonded, "S1"):
                        tofind = True
                        break
            if tofind:
                break
    return tofind


@gaff.Add_Rule("n")
def _rule_n(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3"):
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "C3"):
                for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                    if assign.Atom_Judge(bonded_atom_bonded, "O1") or assign.Atom_Judge(bonded_atom_bonded, "S1"):
                        tofind = True
                        break
            if tofind:
                break
    return tofind


@gaff.Add_Rule("nk")
def _rule_nk(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N4") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("nl")
def _rule_nl(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N4") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("n4")
def _rule_n4(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N4")


@gaff.Add_Rule("no")
def _rule_no(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = 0
    if assign.Atom_Judge(i, "N3"):
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "O1"):
                tofind += 1
    return tofind == 2


@gaff.Add_Rule("na")
def _rule_na(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N3") and \
       ("AR1" in assign.atom_marker[i].keys() or "AR2" in assign.atom_marker[i].keys() or "AR3" in
        assign.atom_marker[i].keys())


@gaff.Add_Rule("nm")
def _rule_nm(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3") and "RG3" in assign.atom_marker[i].keys():
        for bonded_atom in assign.bonds[i].keys():
            if ("DB" in assign.atom_marker[bonded_atom].keys() and
                    (assign.Atom_Judge(bonded_atom, "C3") or
                     assign.Atom_Judge(bonded_atom, "N2") or
                     assign.Atom_Judge(bonded_atom, "P2"))):
                tofind = True
                break
            if (("AR1" in assign.atom_marker[bonded_atom].keys() or
                 "AR2" in assign.atom_marker[bonded_atom].keys() or
                 "AR3" in assign.atom_marker[bonded_atom].keys()) and
                    assign.atoms[bonded_atom] in assign.XX):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("nn")
def _rule_nn(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3") and "RG4" in assign.atom_marker[i].keys():
        for bonded_atom in assign.bonds[i].keys():
            if "DB" in assign.atom_marker[bonded_atom].keys() and assign.Atom_Judge(bonded_atom,
                                                                                    ["C3", "N2", "P2"]):
                tofind = True
                break
            if (("AR1" in assign.atom_marker[bonded_atom].keys() or
                 "AR2" in assign.atom_marker[bonded_atom].keys() or
                 "AR3" in assign.atom_marker[bonded_atom].keys()) and
                    assign.atoms[bonded_atom] in assign.XX):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("nh")
def _rule_nh(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "N3"):
        for bonded_atom in assign.bonds[i].keys():
            if "DB" in assign.atom_marker[bonded_atom].keys() and assign.Atom_Judge(bonded_atom,
                                                                                    ["C3", "N2", "P2"]):
                tofind = True
                break
            if (("AR1" in assign.atom_marker[bonded_atom].keys() or
                 "AR2" in assign.atom_marker[bonded_atom].keys() or
                 "AR3" in assign.atom_marker[bonded_atom].keys()) and
                    assign.atoms[bonded_atom] in assign.XX):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("np")
def _rule_np(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N3") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("nq")
def _rule_nq(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N3") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("n3")
def _rule_n3(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N3")


@gaff.Add_Rule("nb")
def _rule_nb(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N2") and "AR1" in assign.atom_marker[i].keys()


@gaff.Add_Rule("nd")
def _rule_nd(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "N2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
         ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
            if assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]):
                for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                    if assign.Atom_Judge(bonded_atom_bonded, ["C3", "C2", "N2", "P2"]):
                        tofind = True
                        break
                if tofind:
                    break
    if tofind:
        tofind = _single_double_name(i, assign, "nc", "nd")
    return tofind


@gaff.Add_Rule("nc")
def _rule_nc(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "N2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
         ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
            if assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]):
                for bonded_atom_bonded in assign.bonds[bonded_atom].keys():
                    if assign.Atom_Judge(bonded_atom_bonded, ["C3", "C2", "N2", "P2"]):
                        tofind = True
                        break
                if tofind:
                    break
    return tofind


@gaff.Add_Rule("nf")
def _rule_nf(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "N2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    if tofind:
        tofind = _single_double_name(i, assign, "ne", "nf")
    return tofind


@gaff.Add_Rule("ne")
def _rule_ne(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "N2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "C2", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("n1")
def _rule_n1(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N1") or (assign.Atom_Judge(i, "N2") and
                                          (("sb" in assign.atom_marker[i] and "tb" in assign.atom_marker[i]) or
                                           (assign.atom_marker[i].get("db") == 2)))


@gaff.Add_Rule("n2")
def _rule_n2(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "N2")


@gaff.Add_Rule("s")
def _rule_s(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S1")


@gaff.Add_Rule("s2")
def _rule_s2(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S2") and ("DB" in assign.atom_marker[i] or "TB" in assign.atom_marker[i])


@gaff.Add_Rule("sh")
def _rule_sh(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "S2"):
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "H1"):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("sp")
def _rule_sp(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S2") and "RG3" in assign.atom_marker[i].keys()


@gaff.Add_Rule("sq")
def _rule_sq(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S2") and "RG4" in assign.atom_marker[i].keys()


@gaff.Add_Rule("ss")
def _rule_ss(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S2")


@gaff.Add_Rule("sx")
def _rule_sx(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "S3") and "db" in assign.atom_marker[i].keys()
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("s4")
def _rule_s4(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S3")


@gaff.Add_Rule("sy")
def _rule_sy(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "S4") and "db" in assign.atom_marker[i].keys()
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("s6")
def _rule_s6(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "S4") or assign.Atom_Judge(i, "S5") or assign.Atom_Judge(i, "S6")


@gaff.Add_Rule("pd")
def _rule_pd(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "P2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
             ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    if tofind:
        tofind = _single_double_name(i, assign, "pc", "pd")
    return tofind


@gaff.Add_Rule("pb")
def _rule_pb(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "P2") and "AR1" in assign.atom_marker[i].keys()


@gaff.Add_Rule("pc")
def _rule_pc(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "P2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i] and \
         ("AR2" in assign.atom_marker[i] or "AR3" in assign.atom_marker[i])
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("pf")
def _rule_pf(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "P2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    if tofind:
        tofind = _single_double_name(i, assign, "pe", "pf")
    return tofind


@gaff.Add_Rule("pe")
def _rule_pe(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = assign.Atom_Judge(i, "P2") and "sb" in assign.atom_marker[i] and "db" in assign.atom_marker[i]
    if tofind:
        tofind = False
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("p2")
def _rule_p2(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "P1") or assign.Atom_Judge(i, "P2")


@gaff.Add_Rule("px")
def _rule_px(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "P3") and "db" in assign.atom_marker[i].keys():
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("p4")
def _rule_p4(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "P3") and "db" in assign.atom_marker[i].keys():
        for bonded_atom in assign.bonds[i].keys():
            if assign.Atom_Judge(bonded_atom, "O1") or assign.Atom_Judge(bonded_atom, "S1"):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("p3")
def _rule_p3(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "P3")


@gaff.Add_Rule("py")
def _rule_py(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    tofind = False
    if assign.Atom_Judge(i, "P4") and "db" in assign.atom_marker[i].keys():
        for bonded_atom, bond_order in assign.bonds[i].items():
            if bond_order == 1 and (assign.Atom_Judge(bonded_atom, ["C3", "N2", "P2"]) or
                                    (assign.Atom_Judge(bonded_atom, ["S3", "S4", "P3", "P4"]) and "db" in
                                     assign.atom_marker[bonded_atom])):
                tofind = True
                break
    return tofind


@gaff.Add_Rule("p5")
def _rule_p5(i, assign):
    """

    :param i:
    :param assign:
    :return:
    """
    return assign.Atom_Judge(i, "P4") or assign.Atom_Judge(i, "P5") or assign.Atom_Judge(i, "P6")


def parmchk2_gaff(ifname, ofname, direct_load=True, keep=True):
    """
    This **function** is to do parmchk2 for gaff

    :param ifname: a string of mol2 file name, a ResidueType, Residue or Molecule instance
    :param ofname: the output file name
    :param direct_load: directly load the output file after writing it
    :param keep: do not delete the output file after loading it
    :return: None
    """
    import XpongeLib as xlib
    datapath = os.path.split(xlib.__file__)[0]
    if isinstance(ifname, AbstractMolecule):
        Save_Mol2(ifname, "temp.mol2")
        ifname = "temp.mol2"
    parmchk2_func = getattr(xlib, "_parmchk2")
    parmchk2_func(ifname, "mol2", ofname, datapath, 0, 1, 1)
    if direct_load:
        amber.load_parameters_from_frcmod(ofname, prefix=False)
    if not keep:
        os.remove(ofname)
    if os.path.exists("temp.mol2"):
        os.remove("temp.mol2")


set_real_global_variable("parmchk2_gaff", parmchk2_gaff)

Xprint("""Reference for gaff:
  Wang, J., Wolf, R.M., Caldwell, J.W., Kollman, P.A. and Case, D.A.
    Development and testing of a general amber force field.
    Journal of Computational Chemistry 2004 25, 1157-1174
    DOI: 10.1002/jcc.20035
""")
# pylint:disable=undefined-variable



"""
This **package** sets the tips3p configuration of charmm27 force field
"""
from ...helper import source, Xprint, set_real_global_variable

source("....")
source("...charmm27")

AtomType.New_From_String(
    """
    name mass    charge[e]  LJtype
    HW   1.008    0.417       HW
    OW   16      -0.834       OW
    """)

bond_base.BondType.New_From_String(r"""
name   k[kcal/mol·A^-2]   b[A]
OW-HW  553                0.9572
HW-HW  553                1.5136
""")

ub_angle_base.UreyBradleyType.New_From_String(r"""
name      k   kUB  
HW-OW-HW  0   0
OW-HW-HW  0   0
""")

lj_base.LJType.New_From_String(r"""
name    sigma[nm]   epsilon[kJ/mol]
OW-OW   0.315057422683    0.6363864
HW-HW   0.0400013524445 0.192464
""")

TIPS3P = load_mol2(os.path.join(CHARMM27_DATA_DIR, "tip3p.mol2"), as_template=True)

load_mol2(os.path.join(CHARMM27_DATA_DIR, "atomic_ions.mol2"), as_template=True)

set_real_global_variable("WAT", TIPS3P)

Xprint("""Reference for CHARMM modified tip3p:
  to do
""")
# pylint:disable=undefined-variable



"""
This **module** set the basic configuration for OL15 (DNA)
"""
from ...helper import source, Xprint

source("....")
amber = source("...amber")

amber.load_parameters_from_parmdat("parm10.dat")
amber.load_parameters_from_frcmod("OL15.frcmod")

load_mol2(os.path.join(AMBER_DATA_DIR, "ol15.mol2"), as_template=True)

for i in "ATCG":
    i = "D" + i
    res = ResidueType.get_type(i)
    res5 = ResidueType.get_type(i + "5")
    res3 = ResidueType.get_type(i + "3")
    res.tail = "O3'"
    res.tail_next = "C3'"
    res5.tail = "O3'"
    res5.tail_next = "C3'"

    res.head = "P"
    res.head_next = "OP2"
    res3.head = "P"
    res3.head_next = "OP2"

    res.tail_link_conditions.append({"atoms": ["C3'", "O3'"], "parameter": 120 / 180 * np.pi})
    res.tail_link_conditions.append({"atoms": ["H3'", "C3'", "O3'"], "parameter": -54 / 180 * np.pi})
    res5.tail_link_conditions.append({"atoms": ["C3'", "O3'"], "parameter": 120 / 180 * np.pi})
    res5.tail_link_conditions.append({"atoms": ["H3'", "C3'", "O3'"], "parameter": -54 / 180 * np.pi})

    res.head_link_conditions.append({"atoms": ["OP2", "P"], "parameter": 108 / 180 * np.pi})
    res.head_link_conditions.append({"atoms": ["O5'", "OP2", "P"], "parameter": 113 / 180 * np.pi})
    res3.head_link_conditions.append({"atoms": ["OP2", "P"], "parameter": 108 / 180 * np.pi})
    res3.head_link_conditions.append({"atoms": ["O5'", "OP2", "P"], "parameter": 113 / 180 * np.pi})

    GlobalSetting.Add_PDB_Residue_Name_Mapping("head", res.name, res5.name)
    GlobalSetting.Add_PDB_Residue_Name_Mapping("tail", res.name, res3.name)

Xprint("""Reference for OL15:
  Marie Zgarbová, Jiří Šponer, Michal Otyepka, Thomas E Cheatham 3rd, Rodrigo Galindo-Murillo, and Petr Jurečka
    Refinement of the Sugar-Phosphate Backbone Torsion Beta for AMBER Force Fields Improves the Description of 
      Z- and B-DNA
    Journal of Chemical Theory and Computation 2015 11 (8), 5723-5736
    DOI: 10.1021/acs.jctc.5b00716
""")



"""
This **module** is the basic setting for the force field format of 3-parameter non bonded 1-4 interactions
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, GlobalSetting, set_global_alternative_names

# pylint: disable=invalid-name
NB14Type = Generate_New_Bonded_Force_Type("nb14_extra", "1-4",
                                          {"epsilon": float, "rmin": float, "sigma": float, "A": float, "B": float,
                                           "kee": float, "nb14_ee_factor": float}, False)

NB14Type.Set_Property_Unit("rmin", "distance", "A")
NB14Type.Set_Property_Unit("sigma", "distance", "A")
NB14Type.Set_Property_Unit("epsilon", "energy", "kcal/mol")
NB14Type.Set_Property_Unit("A", "energy·distance^6", "kcal/mol·A^6")
NB14Type.Set_Property_Unit("B", "energy·distance^12", "kcal/mol·A^12")

NB14Type.topology_matrix = [[1, -4],
                            [1, 1]]


def get_nb14_extra_lj(atom1, atom2):
    """
    This **function** is used to get the LJ parameters for the extra nb14 interactions

    :param atom1: the Atom instance
    :param atom2: the Atom instance
    :return: the A and B coefficients of LJ
    """
    from . import lj_base
    lj_type = lj_base.LJType

    a = 0
    b = 0

    lj_i = lj_type.get_type(atom1.LJtype + "-" + atom1.LJtype)
    lj_j = lj_type.get_type(atom2.LJtype + "-" + atom2.LJtype)
    finded = False
    findnames = [atom1.LJtype + "-" + atom2.LJtype,
                 atom2.LJtype + "-" + atom1.LJtype]

    for findname in findnames:
        if findname in lj_type.get_all_types():
            finded = True
            lj_ij = lj_type.get_type(findname)
            a = lj_type.combining_method_A(lj_ij.epsilon, lj_ij.rmin, lj_ij.epsilon, lj_ij.rmin)
            b = lj_type.combining_method_B(lj_ij.epsilon, lj_ij.rmin, lj_ij.epsilon, lj_ij.rmin)
            break
    if not finded:
        a = lj_type.combining_method_A(lj_i.epsilon, lj_i.rmin, lj_j.epsilon, lj_j.rmin)
        b = lj_type.combining_method_B(lj_i.epsilon, lj_i.rmin, lj_j.epsilon, lj_j.rmin)
    return a, b


def exclude_to_nb14_extra(molecule, atom1, atom2):
    """
    This **function** is used to calculate nb14_extra instead of non-bonded interactions for atom1 and atom2

    :param molecule: the Molecule instance
    :param atom1: the Atom instance
    :param atom2: the Atom instance
    :return: None
    """
    new_force = NB14Type.entity([atom1, atom2], NB14Type.get_type("UNKNOWNS"))
    a, b = Get_NB14EXTRA_AB(new_force.atoms[0], new_force.atoms[1])
    new_force.A = nb14_bond.kLJ * a
    new_force.B = nb14_bond.kLJ * b
    new_force.kee = nb14_bond.kee
    atom1.Extra_Exclude_Atom(atom2)

    molecule.Add_Bonded_Force(new_force)


def nb14_to_nb14_extra(molecule):
    """
    This **function** is used to convert nb14 to nb14_extra

    :param molecule: the Molecule instance
    :return: None
    """
    # 处理nb14
    # A、B中的nb14全部变为nb14_extra
    while molecule.bonded_forces.get("nb14", []):
        nb14_bond = molecule.bonded_forces["nb14"].pop()
        new_force = NB14Type.entity(nb14_bond.atoms, NB14Type.get_type("UNKNOWNS"), nb14_bond.name)

        a, b = get_nb14_extra_lj(new_force.atoms[0], new_force.atoms[1])

        new_force.A = nb14_bond.kLJ * a
        new_force.B = nb14_bond.kLJ * b
        new_force.kee = nb14_bond.kee
        new_force.nb14_ee_factor = nb14_bond.kee * new_force.atoms[0].charge * new_force.atoms[1].charge
        molecule.Add_Bonded_Force(new_force)


@GlobalSetting.Add_Unit_Transfer_Function(NB14Type)
def _lj_unit_transfer(self):
    """
    This **function** is used to transfer the units of lj

    :param self:
    :return:
    """
    if self.sigma is not None:
        self.rmin = self.sigma * (4 ** (1 / 12) / 2)
        self.sigma = None
    if self.rmin is not None and self.epsilon is not None:
        self.A = self.epsilon * (2 * self.rmin) ** 12
        self.B = self.epsilon * 2 * ((2 * self.rmin) ** 6)
        self.rmin = None
        self.epsilon = None


@Molecule.Set_Save_SPONGE_Input("nb14_extra")
def write_nb14(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    bonds = []
    for bond in self.bonded_forces.get("nb14_extra", []):
        order = list(range(2))
        if bond.A != 0 or bond.B != 0 or bond.kee != 0:
            if self.atom_index[bond.atoms[order[0]]] > self.atom_index[bond.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order

            bonds.append("%d %d %.6e %.6e %.6e" % (self.atom_index[bond.atoms[temp_order[0]]],
                                                   self.atom_index[bond.atoms[temp_order[1]]],
                                                   bond.A, bond.B, bond.kee))

    if bonds:
        towrite = "%d\n" % len(bonds)
        bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
        towrite += "\n".join(bonds)

        return towrite
    return None

set_global_alternative_names()
